//------------------------------------
//  (c) Reliable Software, 1996, 97
//------------------------------------

void ShaderSet(Shader *S, int ShaderType, RAY *incident, VISIBLEOBJECT *obj)
{
  VECTOR light;

  S->Type 		  = ShaderType;
  S->IncidentRay  = incident;
  S->Depth 		  = incident->Depth;
  S->Prop 		  = &obj->OpticalProperty;
  S->ShadowRay 	  = NULL;
  S->ReflectedRay = NULL;
  S->RefractedRay = NULL;

  S->Intersect = RayIntersect(incident);
  S->Normal    = ObjectNormal(obj, &S->Intersect);

  S->Cosine = GetDotProductOfVectors(&S->Normal, &incident->Direction);
  // Vector pointing towards light
  SetVectorFromPoints(&light, &LightPoint, &S->Intersect);
  NormalizeVector(&light);
  S->Light_cos = GetDotProductOfVectors(&light, &S->Normal);

  // Shadow ray towards the source of light
  if (S->Prop->Diffusion > 0.0 && S->Light_cos > 0.0){

    S->ShadowRay = (RAY*)HeapAlloc(g_hHeap, 0, sizeof(RAY));
	//S->ShadowRay = (RAY*)malloc(sizeof(RAY));
    RaySet(S->ShadowRay, rtShadow, &S->Intersect, &light, S->Depth+1, 0);
  }

  // Reflection
  if (S->Prop->Reflectiveness > 0.0){
    VECTOR ref;
    ref = ShaderReflect(S);

    S->ReflectedRay = (RAY*)HeapAlloc(g_hHeap, 0, sizeof(RAY));
    //S->ReflectedRay = (RAY*)malloc(sizeof(RAY));
    RaySet(S->ReflectedRay, rtNormal, &S->Intersect, &ref, S->Depth+1, 0);
  }

  // Refraction
  if (S->Prop->Transparency > 0.0){
    VECTOR ref;
    ref = ShaderRefract(S);
    if (IsVectorNonZero(&ref)){

      S->RefractedRay = (RAY*)HeapAlloc(g_hHeap, 0, sizeof(RAY));
      //S->RefractedRay = (RAY*)malloc(sizeof(RAY));
	  RaySet(S->RefractedRay, rtRefracted, &S->Intersect, &ref, S->Depth+1, S->Prop->Refractiveness);
    }
  }

  if (ShaderType == SHADER_PLANE)
	  ObjectGetPixel(obj, &S->Intersect, &S->Pixel);
}

void ShaderFree(Shader *S)
{
  if (S->ReflectedRay) HeapFree(g_hHeap, 0, S->ReflectedRay);
  if (S->RefractedRay) HeapFree(g_hHeap, 0, S->RefractedRay);
  if (S->ShadowRay)    HeapFree(g_hHeap, 0, S->ShadowRay);
//  if (S->ReflectedRay) free(S->ReflectedRay);
//  if (S->RefractedRay) free(S->RefractedRay);
//  if (S->ShadowRay)    free(S->ShadowRay);
}

void ShaderGetColor(Shader *S,COLOR *final_color)
{
  // I = Ia + kd ( N . L ) + ks S + kt T

  COLOR diffuse = {0,0,0}, specular = {0,0,0}, refr = {0,0,0};


 if (S->ShadowRay){
    RayHit(S->ShadowRay, &diffuse);
    MultiplyColor(&diffuse, S->ShadowRay->Attenuation);
    MultiplyColor(&diffuse, S->Light_cos);
  }

  MultiplyColor(&diffuse, S->Prop->Diffusion);
  AddColorToColor(&diffuse, &Ambient);
  MultiplyColorByColor(&diffuse, &S->Prop->Col);
  *final_color = diffuse;

  // Reflection
  if (S->ReflectedRay){
    RayHit(S->ReflectedRay, &specular);
  }

  MultiplyColor(&specular, S->Prop->Reflectiveness);
  AddColorToColor(final_color, &specular);

  // Refraction
  if (S->RefractedRay){
	 RayHit(S->RefractedRay, &refr);
  }

  MultiplyColor(&refr, S->Prop->Transparency);
  AddColorToColor(final_color, &refr);

  if (S->Type == SHADER_PLANE)
	  MultiplyColorByColor(final_color, &S->Pixel);

}

VECTOR ShaderReflect(Shader *S)
{
  // refl = incid - 2 * ( normal . incid ) normal
  VECTOR result;

  result = S->Normal;
  MultiplyVector(&result, -2 * S->Cosine);
  AddVectorToVector(&result, &S->IncidentRay->Direction);
  return result;
}

VECTOR ShaderRefract(Shader *S)
{
  // Snell's law:   ki * sin ( angle(inc, n) ) = kr * sin ( angle(r * n) )
  // all vectors normalized
  //    ki * cross( inc, n ) =  kr * cross ( r, n )
  //    r = a n + b inc
  //    cross ( ( a n + b inc ), n ) = b cross ( inc, n )
  //    (ki - b * kr ) cross ( inc, n ) = 0
  //    b = ki/kr
  //      r normalized
  //    r . r = 1
  //    ( a n + b inc )^2 = 1
  //    a^2 + 2 a b ( n . inc) + b^2 = 1
  //    a^2 + 2 b cos a + b^2 - 1 = 0
  //    disc = 4 (b^2 cos^2 - b^2 + 1)
  //

  // if cosine negative we are entering the object (from vacuum)
  // ki = 1, kr = refr, b = 1/refr
  // else ki = refr, kr = 1, b = refr
  double refr,disc_2;

  if (S->Cosine >= 0.0)
      refr = S->Prop->Refractiveness;
  else if (S->Prop->Refractiveness > 0.0)
      refr = 1.0 / S->Prop->Refractiveness;

  disc_2 = refr * refr * (S->Cosine * S->Cosine - 1) + 1;

  if (disc_2 > 0.0){
    double discr,alpha;
    VECTOR result,n;

    discr = sqrt (disc_2);
    // a = - b cos - disc_2
    alpha = -refr*S->Cosine;
    if (S->Cosine < 0.0)
      alpha -= discr;
    else
      alpha += discr;

    // r = a n + b inc
    result = S->IncidentRay->Direction;

    MultiplyVector(&result, refr);

    n = S->Normal;
    MultiplyVector(&n, alpha);
    AddVectorToVector(&result, &n);
    return result;
  }else{ // total internal reflection (fake)
    VECTOR Temp = {0, 0, 0};
    return Temp;
  }
}


